/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* program: cavput.c
 * purpose: vector CA puts from command line
 * M. Borland, 1995.
 */
#include "mdb.h"
#include "scan.h"
#include "match_string.h"
#include "SDDS.h"
#include "pvMultList.h"
#include <stdlib.h>
#include <time.h>
/*#include <tsDefs.h>*/
#include <cadef.h>
#include <ezca.h>

#define CLO_LIST 0
#define CLO_RANGE 1
#define CLO_PENDIOTIME 2
#define CLO_DRYRUN 3
/* obsolete options for commandline compatbility */
#define CLO_EZCATIMING 4
#define CLO_NOGROUPS 5
#define CLO_DELTAMODE 6
#define CLO_NUMERICAL 7
#define CLO_BLUNDERAHEAD 8
#define COMMANDLINE_OPTIONS 9
char *commandline_option[COMMANDLINE_OPTIONS] = {
  "list", "range", "pendiotime", "dryrun", "ezcatiming", "nogroups",
  "deltamode", "numerical", "blunderahead",
} ;

static char *USAGE =   "cavput [-list=<string>[=<value>][,<string>[=<value>]...]]\n\
[-range=begin=<integer>,end=<integer>[,format=<string>][,interval=<integer>]]\n\
[-pendIoTime=<seconds>] [-deltaMode[=factor=<value>]]\n\
[-numerical] [-blunderAhead[=silently]]\n\n\
-list           specifies PV name string components\n\
-range          specifies range of integers and format string\n\
-pendIoTime     specifies maximum time to wait for connections and\n\
                return values.  Default is 1.0s \n\
-dryRun         specifies showing PV names and values but\n\
                not sending to IOCs.\n\
-deltaMode      specifies that values are deltas from present values in\n\
                PVs.\n\
-numerical      forces conversion of the value to a number and sending as\n\
                a number.  Default for -deltaMode.\n\
-blunderAhead   specifies that the program should blunder ahead with all the\n\
                puts it can do, even if some PVs don't connect.\n\n\
Program by Michael Borland, ANL/APS (This is version 5, September 1996.)\n";

/* example of usage:
 # set all SR quads to zero
 cavput -list=S -range=begin=1,end=40,format=%ld 
    -list=AQ,BQ -range=begin=1,end=5 -list=:CurrentAI=0
 # set slow beam history for S1A bpms and 4080 samples
 cavput -list=S:bpm -range=begin=1,end=40,format=%ld -list=:SlowBh:
   -list=MI=0,sizeAO=4080,modeBO=1,enableBO=1
 */

#define DO_BLUNDERAHEAD   0x0001UL
#define BLUNDER_SILENTLY  0x0002UL

int main(int argc, char **argv)
{
  long beginRange, endRange, rangeInterval, dryRun;
  long PVs, j, i_arg, status, numerical;
  unsigned long blunderAhead;
  PV_VALUE *PVvalue;
  TERM_LIST *List;
  unsigned long flags, dummyFlags;
  SCANNED_ARG *s_arg;
  char *ptr, *rangeFormat;
  double pendIOTime;
  chid *channelID;
  short deltaMode;
  double *presentValue, deltaFactor;
  
  argc = scanargs(&s_arg, argc, argv);
  if (argc<2) 
    bomb(NULL, USAGE);

  PVvalue = NULL;
  List = NULL;
  PVs = dryRun = 0;
  pendIOTime = 1.0;
  deltaMode = numerical = 0;
  deltaFactor = 1;
  blunderAhead = 0;
  
  for (i_arg=1; i_arg<argc; i_arg++) {
    if (s_arg[i_arg].arg_type==OPTION) {
      delete_chars(s_arg[i_arg].list[0], "_");
      switch (match_string(s_arg[i_arg].list[0], commandline_option, COMMANDLINE_OPTIONS, 0)) {
      case CLO_LIST:
        if (s_arg[i_arg].n_items<2)
          SDDS_Bomb("invalid -list syntax (cavput)");
        List = trealloc(List, sizeof(*List)*(s_arg[i_arg].n_items-1));
        for (j=1; j<s_arg[i_arg].n_items; j++) {
          List[j-1].flags = 0;
          if ((ptr=strchr(s_arg[i_arg].list[j], '=')))
            *ptr++ = 0;
          List[j-1].string = s_arg[i_arg].list[j];
          if (ptr) {
            List[j-1].flags |= VALUE_GIVEN;
            List[j-1].value = ptr;
          }
        }
        multiplyWithList(&PVvalue, &PVs, List, s_arg[i_arg].n_items-1);
        break;
      case CLO_RANGE:
        s_arg[i_arg].n_items --;
        rangeFormat = NULL;
        rangeInterval = 1;
        if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "begin", SDDS_LONG, &beginRange, 1, BEGIN_GIVEN,
                          "end", SDDS_LONG, &endRange, 1, END_GIVEN,
                          "interval", SDDS_LONG, &rangeInterval, 1, INTERVAL_GIVEN,
                          "format", SDDS_STRING, &rangeFormat, 1, FORMAT_GIVEN,
                          NULL) ||
            !(flags&BEGIN_GIVEN) || !(flags&END_GIVEN) || beginRange>endRange ||
            (flags&FORMAT_GIVEN && SDDS_StringIsBlank(rangeFormat)))
          SDDS_Bomb("invalid -range syntax/values");
        if (!rangeFormat)
          rangeFormat = "%ld";
        multiplyWithRange(&PVvalue, &PVs, beginRange, endRange, rangeInterval, rangeFormat);
        break;
      case CLO_PENDIOTIME:
        if (s_arg[i_arg].n_items!=2)
          bomb("wrong number of items for -pendIoTime", NULL);
        if (sscanf(s_arg[i_arg].list[1], "%lf", &pendIOTime)!=1 ||
            pendIOTime<=0)
          SDDS_Bomb("invalid -pendIoTime value (cavget)");
        break;
      case CLO_DRYRUN:
        dryRun = 1;
        break;
      case CLO_EZCATIMING:
        fprintf(stderr, "warning (cavput): -ezcaTiming option is obsolete. Use -pendIoTime\n");
        break;
      case CLO_NOGROUPS:
        fprintf(stderr, "warning (cavput): -noGroups option is obsolete.\n");
        break;
      case CLO_DELTAMODE:
        deltaMode = 1;
        s_arg[i_arg].n_items--;
        if (!scanItemList(&dummyFlags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "factor", SDDS_DOUBLE, &deltaFactor, 1, 0,
                          NULL))
          SDDS_Bomb("invalid -deltaMode syntax/values");
        break;
      case CLO_NUMERICAL:
        numerical = 1;
        break;
      case CLO_BLUNDERAHEAD:
        s_arg[i_arg].n_items--;
        if (!scanItemList(&blunderAhead, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                          "silently", -1, NULL, 0, BLUNDER_SILENTLY,
                          NULL))
          SDDS_Bomb("invalid -blunderAhead syntax/values");
        blunderAhead |= DO_BLUNDERAHEAD;
        break;
      default:
        bomb("unknown option (cavput)", NULL);
        break;
      }
    }
    else
      bomb("unknown option", NULL);
  }

  for (j=0; j<PVs; j++)
    if (!PVvalue[j].value) {
      fprintf(stderr, "Error: no value given for %s\n", PVvalue[j].name);
      exit(1);
    }
  
  if (dryRun) {
    for (j=0; j<PVs; j++)
      printf("%32s %s\n", PVvalue[j].name, PVvalue[j].value);
  }
  else {
    ca_task_initialize();
    if (!(channelID = (chid*)malloc(sizeof(*channelID)*PVs)))
      bomb("memory allocation failure (cavput)", NULL);
    for (j=0; j<PVs; j++)
      if (ca_search(PVvalue[j].name, channelID+j)!=ECA_NORMAL) {
        fprintf(stderr, "error (cavput): problem doing search for %s\n", 
                PVvalue[j].name);
        exit(1);
      }
    ca_pend_io(pendIOTime);
    for (j=0; j<PVs; j++) {
      if (ca_state(channelID[j])!=cs_conn) {
        if (!(blunderAhead&BLUNDER_SILENTLY))
          fprintf(stderr, "error (cavput): no connection for %s\n",
                  PVvalue[j].name);
        if (!blunderAhead)
          exit(1);
      }
    }
    if (deltaMode) {
      if (!(presentValue = (double*)malloc(sizeof(*presentValue)*PVs))) {
        fprintf(stderr, "error (cavput): memory allocation failure\n");
        exit(1);
      }
      for (j=0; j<PVs; j++) {
        presentValue[j] = DBL_MAX;
        if (ca_state(channelID[j])!=cs_conn)
          continue;
        if (ca_get(DBR_DOUBLE, channelID[j], presentValue+j)!=ECA_NORMAL) {
          fprintf(stderr, "error (cavput): problem doing get for %s\n",
                  PVvalue[j].name);
          exit(1);
        }
      }
      ca_pend_io(pendIOTime);
      for (j=0; j<PVs; j++) {
        if (ca_state(channelID[j])!=cs_conn)
          continue;
        if (presentValue[j]==DBL_MAX) {
          fprintf(stderr, "error (cavput): no value returned in time for %s\n",
                  PVvalue[j].name);
          exit(1);
        }
        if (sscanf(PVvalue[j].value, "%le", &PVvalue[j].numericalValue)!=1) {
          fprintf(stderr, "error (cavput): value (%s) for %s is not numerical--can't use -deltaMode\n",
                  PVvalue[j].value, PVvalue[j].name);
          exit(1);
        }
        PVvalue[j].numericalValue = PVvalue[j].numericalValue*deltaFactor + presentValue[j];
        /*                fprintf(stderr, "%s now at %e, ", PVvalue[j].name, presentValue[j]); */
        /*                fprintf(stderr, "will go to %e\n", PVvalue[j].numericalValue); */
      }
      for (j=0; j<PVs; j++) {
        if (ca_state(channelID[j])!=cs_conn)
          continue;
        if (ca_put(DBR_DOUBLE, channelID[j], &PVvalue[j].numericalValue)!=ECA_NORMAL) {
          if (!(blunderAhead&BLUNDER_SILENTLY))
            fprintf(stderr, "error (cavput): problem doing put for %s\n",
                    PVvalue[j].name);
          if (!blunderAhead)
            exit(1);
        }
      }
    }
    else {
      for (j=0; j<PVs; j++) {
        if (ca_state(channelID[j])!=cs_conn)
          continue;
        if (!numerical) {
          if (ca_put(DBR_STRING, channelID[j], PVvalue[j].value)
              !=ECA_NORMAL) {
            if (!(blunderAhead&BLUNDER_SILENTLY))
              fprintf(stderr, "error (cavput): problem doing put for %s\n",
                      PVvalue[j].name);
            if (!blunderAhead)
              exit(1);
          }
        }
        else {
          if (sscanf(PVvalue[j].value, "%le", &PVvalue[j].numericalValue)!=1) {
            fprintf(stderr, 
                    "error (cavput): value (%s) for %s is not scannable as a number---can't use -numerical.\n",
                    PVvalue[j].value, PVvalue[j].name);
            exit(1);
          }
          if (ca_put(DBR_DOUBLE, channelID[j], &PVvalue[j].numericalValue)!=ECA_NORMAL) {
            if (!(blunderAhead&BLUNDER_SILENTLY))
              fprintf(stderr, "error (cavput): problem doing put for %s\n",
                      PVvalue[j].name);
            if (!blunderAhead)
              exit(1);
          }
        }
      }
    }
    if ((status=ca_pend_io(pendIOTime))!=ECA_NORMAL) {
      fprintf(stderr, "Problem processing one or more puts (cavput):\n%s\n",
              ca_message(status));
      exit(1);
    }
  }
  for (j=0;j<PVs;j++) {
    if (PVvalue[j].name) free(PVvalue[j].name);
    if (PVvalue[j].value) free(PVvalue[j].value);
    PVvalue[j].name=PVvalue[j].value=NULL;
  }
  free(PVvalue);
  free_scanargs(&s_arg,argc);
  if (List)
    free(List);
  if (channelID) free(channelID);
  
  return(0);
}
