/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* program: cawait.c
 * purpose: waits for a set of PVs to take on values satisfying
 *          given criteria
 * sections lifted from rpn code.
 * M. Borland, 1995.
 */
#include "mdb.h"
#include "scan.h"
#include "match_string.h"
#include "SDDS.h"
#include <stdlib.h>
#include <time.h>
/*#include <tsDefs.h>*/
#include <cadef.h>
#include <sys/types.h>
#if defined(_WIN32)
#include <winsock.h>
#else
#include <sys/time.h>
#endif
#include <stddef.h>

#define CLO_INTERVAL 0
#define CLO_WAITFOR 1
#define CLO_EZCATIMING 2
#define CLO_TIMELIMIT 3
#define CLO_USEMONITOR 4
#define CLO_NOT 5
#define CLO_OR 6
#define CLO_AND 7
#define CLO_REPEAT 8
#define CLO_EMIT 9
#define CLO_PREEVENT 10
#define CLO_POSTEVENT 11
#define CLO_SUBPROCESS 12
#define CLO_ONEVENT 13
#define CLO_ONEND 14
#define CLO_TOTALTIMELIMIT 15
#define CLO_PENDIOTIME 16
#define COMMANDLINE_OPTIONS 17
char *commandline_option[COMMANDLINE_OPTIONS] = {
    "interval", "waitfor", "ezcatiming", "timelimit", "usemonitor",
    "not", "or", "and", "repeat", "emit", "preevent",
    "postevent", "subprocess", "onevent", "onend",
    "totaltimelimit", "pendiotime",
    } ;

static char *USAGE = 
"cawait [-interval=<seconds>] [-timeLimit=<seconds>] [-totalTimeLimit=<seconds>]\n\
-waitFor=<PVname>[,lowerLimit=<value>,upperLimit=<value>][,equalTo=<value>][,sameAs=<string>]\n\
 [,above=<value>][,below=<value>] \n\
 [{-and | -or | -not}]\n\
 [-repeat[=<number>]]\n\
 [-emit=event=<string>[,timeout=<string>][,end=<string>]]\n\
 [-preEvent=<command>]\n\
 [-onEvent=<command>]\n\
 [-postEvent=<command>]\n\
 [-onEnd=<command>]\n\
 [-subprocess=<command>[,event=<string>][,timeout=<string>][,end=<string>]]\n\
 [-pendIOTime=<seconds>]\n\
-interval    specifies interval between checking PVs. Default is 1 s.\n\
-timeLimit   specifies maximum time to wait. Default is infinite.\n\
-totalTimeLimit \n\
             specifies maximum time cawait will run.  Default is infinite.\n\
-waitFor     specifies PV name, plus criterion for ending wait.\n\
             Each occurence pushes a result on the logic stack.\n\
-and, -or, -not\n\
             Performs operations on the logic stack.\n\
-repeat      specifies number of times to wait for the event. Default is\n\
             infinite if -repeat is given with no qualifiers, otherwise the\n\
             default is 1. The event condition must become false to start a new\n\
             event.\n\
-emit        specifies string sto emit to the standard output when the event\n\
             condition occurs, when a timeout occurs, or when the specified\n\
             number of events has been detected.\n\
-preEvent    specifies a command to execute before waiting for the event.\n\
-onEvent     specifies a command to execute when the event occurs.\n\
-postEvent   specifies a command to execute after waiting for the event.\n\
-onEnd       specifies a command to execute after the final event has been detected.\n\
-subprocess  specifies a command to execute on a pipe and optionally the messages to\n\
             pass to the subprocess for events, timeouts, and end-of-run.\n\
-pendIOTime  set time (in seconds) allowed for CA operations.\n\
Program by Michael Borland, ANL/APS (This is version 4, April 1997.)\n";


#define LOWERLIMIT_GIVEN 0x00001UL
#define UPPERLIMIT_GIVEN 0x00002UL
#define EQUALTO_GIVEN    0x00004UL
#define ABOVE_GIVEN      0x00008UL
#define BELOW_GIVEN      0x00010UL
#define SAME_AS_GIVEN    0x00020UL

#define OP_NOT CLO_NOT
#define OP_OR  CLO_OR
#define OP_AND CLO_AND

typedef struct {
    char *PVname;
    chid channelID;
    long channelType;
    double lowerLimit, upperLimit, equalTo;
    double above, below;
    long operations;
    short *operation;
    char *sameAs;
    unsigned long flags;
    double value;
    char stringValue[256];
    } WAITFOR;

typedef struct {
    char *command;
    char *emitOnEvent, *emitOnTimeout, *emitOnEnd;
    FILE *fp;
    } SUBPROCESS;

#define EMIT_EVENT_GIVEN   0x0001UL
#define EMIT_TIMEOUT_GIVEN 0x0002UL
#define EMIT_END_GIVEN     0x0004UL
#define EVENT 0
#define TIMEOUT 1
#define LOGIC_ERROR 2

long waitForEvent(WAITFOR *waitFor, long waitFors, double timeLimit, double interval, 
                  long invert);
long stack_test(long stack_ptr, long numneeded, char *stackname, char *caller);
long pop_log(long *logical);
long push_log(long logical);
void log_and(void);
void log_or(void);
void log_not(void);
void clear_stack();

double getTimeInSecs()
{
#if !defined(SOLARIS)
#include <sys/types.h>
#include <sys/timeb.h>
    struct timeb theTime;
    ftime(&theTime);
    return ((double)theTime.time) + ((double)theTime.millitm)/1000.0;
#else
#include <time.h>
    struct timespec tp;
    clock_gettime(CLOCK_REALTIME, &tp);
    return ((double)tp.tv_sec) + ((double)tp.tv_nsec)/1e9;
#endif
    }

double pendIOTime;

int main(int argc, char **argv)
{
    long i_arg, ezcaRetries, j, terminateLoop;
    unsigned long flags;
    SCANNED_ARG *s_arg;
    double ezcaTimeout;
    WAITFOR *waitFor;
    long waitFors;
    double interval, timeLimit, totalTimeLimit, startTime;
    long useMonitor, code=0, repeats, eventEndPending;
    char *emitOnEvent, *emitOnTimeout, *emitOnEnd;
    char **preEvent, **postEvent, **onEvent, **onEnd;
    SUBPROCESS *subprocess;
    long preEvents, postEvents, onEvents, onEnds, subprocesses;
    
    SDDS_RegisterProgramName(argv[0]);
    argc = scanargs(&s_arg, argc, argv);
    if (argc<2) 
        bomb(NULL, USAGE);

    ezcaTimeout = 0.010;
    ezcaRetries = 1000;
    waitFor = NULL;
    waitFors = 0;
    interval = 0.1;
    timeLimit = totalTimeLimit = -1;
    useMonitor = 0;
    repeats = 1;
    emitOnEvent = emitOnTimeout = emitOnEnd = NULL;
    preEvent = postEvent = onEvent = onEnd = NULL;
    subprocess = NULL;
    preEvents = postEvents = onEvents = onEnds = subprocesses = 0;
    pendIOTime = 10;
    
    for (i_arg=1; i_arg<argc; i_arg++) {
        if (s_arg[i_arg].arg_type==OPTION) {
            delete_chars(s_arg[i_arg].list[0], "_");
            switch (code=match_string(s_arg[i_arg].list[0], commandline_option, COMMANDLINE_OPTIONS, 0)) {
              case CLO_INTERVAL:
                if (s_arg[i_arg].n_items!=2 ||
                    sscanf(s_arg[i_arg].list[1], "%lf", &interval)!=1 ||
                    interval<=0)
                    SDDS_Bomb("invalid -interval syntax/value");
                break;
              case CLO_TIMELIMIT:
                if (s_arg[i_arg].n_items!=2 ||
                    sscanf(s_arg[i_arg].list[1], "%lf", &timeLimit)!=1 ||
                    timeLimit<=0)
                    SDDS_Bomb("invalid -timeLimit syntax/value");
                break;
              case CLO_WAITFOR:
                if (s_arg[i_arg].n_items<3)
                    SDDS_Bomb("invalid -waitFor syntax");
                waitFor = trealloc(waitFor, sizeof(*waitFor)*(waitFors+1));
                waitFor[waitFors].PVname = s_arg[i_arg].list[1];
                s_arg[i_arg].n_items -= 2;
                if (!scanItemList(&waitFor[waitFors].flags, s_arg[i_arg].list+2, &s_arg[i_arg].n_items, 0,
                                    "lowerlimit", SDDS_DOUBLE, &waitFor[waitFors].lowerLimit, 1, LOWERLIMIT_GIVEN,
                                    "upperlimit", SDDS_DOUBLE, &waitFor[waitFors].upperLimit, 1, UPPERLIMIT_GIVEN,
                                    "above", SDDS_DOUBLE, &waitFor[waitFors].above, 1, ABOVE_GIVEN,
                                    "below", SDDS_DOUBLE, &waitFor[waitFors].below, 1, BELOW_GIVEN,
                                    "equal", SDDS_DOUBLE, &waitFor[waitFors].equalTo, 1, EQUALTO_GIVEN,
                                    "sameas", SDDS_STRING, &waitFor[waitFors].sameAs, 1, SAME_AS_GIVEN,
                                    NULL) ||
                    !waitFor[waitFors].flags ||
                    (waitFor[waitFors].flags&LOWERLIMIT_GIVEN && !(waitFor[waitFors].flags&UPPERLIMIT_GIVEN)) ||
                    (!(waitFor[waitFors].flags&LOWERLIMIT_GIVEN) && waitFor[waitFors].flags&UPPERLIMIT_GIVEN) ||
                    (waitFor[waitFors].flags&LOWERLIMIT_GIVEN && 
                     waitFor[waitFors].lowerLimit>=waitFor[waitFors].upperLimit) || 
                    (waitFor[waitFors].flags&SAME_AS_GIVEN && waitFor[waitFors].flags&(~SAME_AS_GIVEN)))
                    SDDS_Bomb("invalid -waitFor syntax/values");
                waitFor[waitFors].operations = 0;
                waitFor[waitFors].operation = NULL;
                waitFors++;
                break;
              case CLO_EZCATIMING:
/*
                if (s_arg[i_arg].n_items!=3)
                    SDDS_Bomb("wrong number of items for -ezcaTiming");
                if (sscanf(s_arg[i_arg].list[1], "%lf", &ezcaTimeout)!=1 ||
                    sscanf(s_arg[i_arg].list[2], "%ld", &ezcaRetries)!=1 ||
                    ezcaTimeout<=0 || ezcaRetries<=0)
                    SDDS_Bomb("invalid -ezcaTiming values");
                ezcaSetTimeout(ezcaTimeout);
                ezcaSetRetryCount(ezcaRetries);
*/
                break;
              case CLO_USEMONITOR:
                useMonitor = 1;
                break;
              case CLO_NOT:
              case CLO_OR:
              case CLO_AND:
                if (waitFors<1 && code==CLO_NOT)
                    SDDS_Bomb("invalid syntax---give a -waitFor option before -not");
                if (waitFors<2 && code!=CLO_NOT)
                    SDDS_Bomb("invalid syntax---give at least two -waitFor options before -or or -and");
                waitFor[waitFors-1].operation 
                    = SDDS_Realloc(waitFor[waitFors-1].operation, 
                                   sizeof(*waitFor[waitFors-1].operation)*(waitFor[waitFors-1].operations+1));
                waitFor[waitFors-1].operation[waitFor[waitFors-1].operations] = code;
                waitFor[waitFors-1].operations++;
                break;
              case CLO_REPEAT:
                if (s_arg[i_arg].n_items==1) 
                    repeats = 0;   /* indefinite number of repeats */
                else if (s_arg[i_arg].n_items!=2 ||
                         sscanf(s_arg[i_arg].list[1], "%ld", &repeats)!=1 ||
                         repeats<1)
                    SDDS_Bomb("invalid -repeats syntax/value");
                break;
              case CLO_EMIT:
                s_arg[i_arg].n_items -= 1;
                if (!scanItemList(&flags, s_arg[i_arg].list+1, &s_arg[i_arg].n_items, 0,
                                    "event", SDDS_STRING, &emitOnEvent, 1, EMIT_EVENT_GIVEN,
                                    "timeout", SDDS_STRING, &emitOnTimeout, 1, EMIT_TIMEOUT_GIVEN,
                                    "end", SDDS_STRING, &emitOnEnd, 1, EMIT_END_GIVEN,
                                    NULL) ||
                    !(flags&EMIT_EVENT_GIVEN))
                    SDDS_Bomb("invalid -emit syntax/values");
                break;
              case CLO_PREEVENT:
                if (s_arg[i_arg].n_items!=2)
                    SDDS_Bomb("invalid -preEvent syntax");
                if (!(preEvent = SDDS_Realloc(preEvent, sizeof(*preEvent)*(preEvents+1))))
                    SDDS_Bomb("allocation failure");
                preEvent[preEvents] = s_arg[i_arg].list[1];
                preEvents++;
                break;
              case CLO_POSTEVENT:
                if (s_arg[i_arg].n_items!=2)
                    SDDS_Bomb("invalid -postEvent syntax");
                if (!(postEvent = SDDS_Realloc(postEvent, sizeof(*postEvent)*(postEvents+1))))
                    SDDS_Bomb("allocation failure");
                postEvent[postEvents] = s_arg[i_arg].list[1];
                postEvents++;
                break;
              case CLO_SUBPROCESS:
                if (s_arg[i_arg].n_items<2)
                    SDDS_Bomb("invalid -subprocess syntax");
                if (!(subprocess = SDDS_Realloc(subprocess, sizeof(*subprocess)*(subprocesses+1))))
                    SDDS_Bomb("allocation failure");
                subprocess[subprocesses].command = s_arg[i_arg].list[1];
                s_arg[i_arg].n_items -= 2;
                if (!scanItemList(&flags, s_arg[i_arg].list+2, &s_arg[i_arg].n_items, 0,
                                    "event", SDDS_STRING, &subprocess[subprocesses].emitOnEvent, 1, EMIT_EVENT_GIVEN,
                                    "timeout", SDDS_STRING, &subprocess[subprocesses].emitOnTimeout, 1, EMIT_TIMEOUT_GIVEN,
                                    "end", SDDS_STRING, &subprocess[subprocesses].emitOnEnd, 1, EMIT_END_GIVEN,
                                    NULL) ||
                    !(flags&EMIT_EVENT_GIVEN))
                    SDDS_Bomb("invalid -subprocess syntax/values");
                subprocess[subprocesses].fp = NULL;
                subprocesses++;
                break;
              case CLO_ONEVENT:
                if (s_arg[i_arg].n_items!=2)
                    SDDS_Bomb("invalid -onEvent syntax");
                if (!(onEvent = SDDS_Realloc(onEvent, sizeof(*onEvent)*(onEvents+1))))
                    SDDS_Bomb("allocation failure");
                onEvent[onEvents] = s_arg[i_arg].list[1];
                onEvents++;
                break;
              case CLO_ONEND:
                if (s_arg[i_arg].n_items!=2)
                    SDDS_Bomb("invalid -onEnd syntax");
                if (!(onEnd = SDDS_Realloc(onEnd, sizeof(*onEnd)*(onEnds+1))))
                    SDDS_Bomb("allocation failure");
                onEnd[onEnds] = s_arg[i_arg].list[1];
                onEnds++;
                break;
              case CLO_TOTALTIMELIMIT:
                if (s_arg[i_arg].n_items!=2 ||
                    sscanf(s_arg[i_arg].list[1], "%lf", &totalTimeLimit)!=1 ||
                    totalTimeLimit<=0)
                    SDDS_Bomb("invalid -totalTimeLimit syntax/value");
                break;
              case CLO_PENDIOTIME:
                if (s_arg[i_arg].n_items!=2)
                  bomb("wrong number of items for -pendIoTime", NULL);
                if (sscanf(s_arg[i_arg].list[1], "%lf", &pendIOTime)!=1 ||
                    pendIOTime<=0)
                  bomb("invalid -pendIoTime value (cawait)", NULL);
                break;
              default:
                SDDS_Bomb("unknown option");
                break;
                }
            }
        else
            SDDS_Bomb("unknown option");
        }

    if (!waitFors)
        SDDS_Bomb("-waitFor option must be given at least once");

    ca_task_initialize();
    
    /* connect to all PVs */
    for (j=0; j<waitFors; j++) {
      if (ca_search(waitFor[j].PVname, &waitFor[j].channelID)!=ECA_NORMAL) {
        fprintf(stderr, "Error (cawait): problem doing seardch for %s\n",
                waitFor[j].PVname);
        exit(1);
      }
    }
    ca_pend_io(pendIOTime);
    for (j=0; j<waitFors; j++) {
      if (ca_state(waitFor[j].channelID)!=cs_conn) {
        fprintf(stderr, "Error (cawait): no connection in time for %s\n",
                waitFor[j].PVname);
        exit(1);
      }
      if (!(waitFor[j].flags&SAME_AS_GIVEN))
        waitFor[j].channelType = DBF_DOUBLE;
      else
        waitFor[j].channelType = DBF_STRING;
    }

    if (useMonitor) {
      /* this option isn't implemented now */
    }
    
    /* initialize the subprocesses */
    for (j=0; j<subprocesses; j++)
      if (!(subprocess[j].fp=popen(subprocess[j].command, "w"))) {
        fprintf(stderr, "Error (cawait): problem starting subprocess with command %s\n",
                subprocess[j].command);
        
        exit(1);
      }

    startTime = getTimeInSecs();
    if (totalTimeLimit>0 && totalTimeLimit<timeLimit)
      timeLimit = totalTimeLimit;
    eventEndPending = 0;  /* indicates that the end of the last event is pending. */
    do {
      terminateLoop = 0;
      if (totalTimeLimit>0 && ((getTimeInSecs() - startTime)>totalTimeLimit)) 
        terminateLoop = 1;

      if (!eventEndPending) {
        if (!terminateLoop) {
          /* execute pre-event commands */
          for (j=0; j<preEvents; j++)
            system(preEvent[j]);

          /* wait for event to become true */
          code = waitForEvent(waitFor, waitFors, timeLimit, interval, 0);

          if (code==LOGIC_ERROR) {
            fputs("logic error---probably too few operators (cawait)\n", stderr);
            exit(1);
          }
          
          /* emit messages to standard output, as requested */
          if (emitOnEvent) {
            switch (code) {
            case EVENT:
              puts(emitOnEvent);
              fflush(stdout);
              break;
            case TIMEOUT:
              if (emitOnTimeout) {
                puts(emitOnTimeout);
                fflush(stdout);
              }
              break;
            default:
              break;
            }
          }
          
          /* emit messages to subprocesses, as requested */
          for (j=0; j<subprocesses; j++) {
            switch (code) {
            case EVENT:
              fputs(subprocess[j].emitOnEvent, subprocess[j].fp);
              fputc('\n', subprocess[j].fp);
              fflush(subprocess[j].fp);
              break;
            case TIMEOUT:
              if (subprocess[j].emitOnTimeout) {
                fputs(subprocess[j].emitOnTimeout, subprocess[j].fp);
                fputc('\n', subprocess[j].fp);
                fflush(subprocess[j].fp);
              }
              break;
            }
          }

          /* execute on-event commands */
          if (code==EVENT)
            for (j=0; j<onEvents; j++)
              system(onEvent[j]);
        }
      }      

      if (repeats==1 || terminateLoop) {
        if (!eventEndPending) {
          /* last repeat */
          if (emitOnEnd) {
            /* emit end-of-run message to stdout */
            puts(emitOnEnd);
            fflush(stdout);
          }
          /* execute end-of-run commands */
          for (j=0; j<onEnds; j++)
            system(onEnd[j]);
          /* emit end-of-run messages to subprocesses */
          for (j=0; j<subprocesses; j++) {
            if (subprocess[j].emitOnEnd) {
              fputs(subprocess[j].emitOnEnd, subprocess[j].fp);
              fputc('\n', subprocess[j].fp);
              fflush(subprocess[j].fp);
            }
          }
        if (terminateLoop || eventEndPending) 
          exit(0);
        /* exit with error if non-event end */
        exit(code==EVENT ? 0 : 1);
        }
      }
      
      /* wait for event to become false */
      code = waitForEvent(waitFor, waitFors, timeLimit, interval, 1);
      if (code==TIMEOUT) {
        eventEndPending = 1;
        fprintf(stderr, "warning: timeout error waiting for event to clear (cawait)\n");
      } else {
        eventEndPending = 0;
        /* execute post-event commands */
        for (j=0; j<postEvents; j++)
          system(postEvent[j]);
      }
    } while (--repeats) ;
    SDDS_Bomb("Abnormal exit from repeats loop--this shouldn't happen");
    return(1);
  }

void sleep_us(unsigned int nusecs) {
   struct timeval tval;

   tval.tv_sec = nusecs / 1000000;
   tval.tv_usec = nusecs % 1000000;
   select(0, NULL, NULL, NULL, &tval);
 }

#define LOGICSTACKSIZE 500

/* stack for logical operations */
short logicstack[LOGICSTACKSIZE];
long lstackptr=0;

long waitForEvent(WAITFOR *waitFor, long waitFors, double timeLimit, double interval, 
                  long invert)
{
    long time0, timeNow;
    long j, k, result, term;

    time(&time0);
    if (timeLimit>0)
        timeLimit += time0;
    while (timeLimit<0 || (time(&timeNow), timeNow<timeLimit)) {
        for (j=0; j<waitFors; j++) {
          /* request data for all PVs */
          if (ca_get(waitFor[j].channelType, waitFor[j].channelID, 
                     (waitFor[j].channelType==DBF_DOUBLE?(void*)&waitFor[j].value:(void*)&waitFor[j].stringValue))!=ECA_NORMAL) {
            fprintf(stderr, "Error (cawait): problem doing put for %s\n",
                    waitFor[j].PVname);
          }
        }
        ca_pend_io(pendIOTime);
        /*
        for (j=0; j<waitFors; j++) {
          if (waitFor[j].channelType==DBF_DOUBLE) {
            fprintf(stderr, "%s = %e\n",
                    waitFor[j].PVname, waitFor[j].value);
          }
          else {
            fprintf(stderr, "%s = %s\n",
                    waitFor[j].PVname, waitFor[j].stringValue);
          }
        }
        */

        clear_stack();
        for (j=0; j<waitFors; j++) {
            term = 1;
            if (waitFor[j].flags&SAME_AS_GIVEN) {
              term = !strcmp(waitFor[j].stringValue, waitFor[j].sameAs);
            }
            else if (waitFor[j].flags&EQUALTO_GIVEN) {
                if (waitFor[j].value!=waitFor[j].equalTo) 
                    term = 0;
                }
            else if (waitFor[j].flags&(LOWERLIMIT_GIVEN+UPPERLIMIT_GIVEN)) {
                if (waitFor[j].value<waitFor[j].lowerLimit ||
                    waitFor[j].value>waitFor[j].upperLimit)
                    term = 0;
                }
            else if (waitFor[j].flags&ABOVE_GIVEN) {
                if (waitFor[j].value<=waitFor[j].above)
                    term = 0;
                }
            else if (waitFor[j].flags&BELOW_GIVEN) {
                if (waitFor[j].value>=waitFor[j].below)
                    term = 0;
                }
            else {
                fprintf(stderr, "no flags set for PV %s--this shouldn't happen\n",
                        waitFor[j].PVname);
                exit(1);
                }
            push_log(term);
            for (k=0; k<waitFor[j].operations; k++) {
                switch (waitFor[j].operation[k]) {
                  case OP_NOT:
                    log_not();
                    break;
                  case OP_OR:
                    log_or();
                    break;
                  case OP_AND:
                    log_and();
                    break;
                  default:
                    SDDS_Bomb("invalid operation seen---this shouldn't happen");
                    break;
                    }
                }
            }
        pop_log(&result);
        if (lstackptr!=0)
            return LOGIC_ERROR;
        if (invert)
            result = !result;
        if (result)
            return EVENT;
        if (interval)
          sleep_us((unsigned int)(1e6*interval));
        }
    return TIMEOUT;
    }

void clear_stack()
{
    lstackptr = 0;
    }

long stack_test(long stack_ptr, long numneeded, char *stackname, char *caller)
{
    if (stack_ptr<numneeded) {
        fprintf(stderr, "too few items on %s stack (%s)\n", stackname, caller);
        return 0;
        }
    return(1);
    }

long pop_log(long *logical)
{
    if (lstackptr<1) {
        fputs("too few items on logical stack (pop_log)\n", stderr);
        exit(1);
        }
    *logical = logicstack[--lstackptr];
    return(1);
    }

long push_log(long logical)
{
    if (lstackptr==LOGICSTACKSIZE) {
        fputs("stack overflow--logical stack size exceeded (push_log)\n", stderr);
        exit(1);
        }
    logicstack[lstackptr++] = logical;
    return(1);
    }


void log_and(void)
{
    if (!stack_test(lstackptr, 2, "logical", "log_and"))
        exit(1);
    logicstack[lstackptr-2] = (logicstack[lstackptr-1] && logicstack[lstackptr-2]);
    lstackptr--;
    }

void log_or(void)
{
    if (!stack_test(lstackptr, 2, "logical", "log_or"))
        exit(1);
    logicstack[lstackptr-2] = (logicstack[lstackptr-1] || logicstack[lstackptr-2]);
    lstackptr--;
    }

void log_not(void)
{
    if (!stack_test(lstackptr, 1, "logical", "log_not"))
        exit(1);
    logicstack[lstackptr-1] = !logicstack[lstackptr-1];
    }

